# -*- encoding: utf-8 -*-
"""
keri.app.cli.commands module

"""
import argparse

from hio.base import doing

import keri

parser = argparse.ArgumentParser(description='Print version of KLI')
parser.set_defaults(handler=lambda args: handler(args))


def handler(args):
    return [doing.doify(version)]


def version(tymth, tock=0.0):
    """ Command line version handler
    """
    _ = (yield tock)

    print(keri.__version__)



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from hio import help
from hio.base import doing

from keri import kering
from keri.app.cli.common import displaying, existing
from keri.core import coring, serdering
from keri.db import dbing, basing
from keri.help import helping
from keri.kering import ConfigurationError

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Revert an unpublished interaction event at the end of a local KEL')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', default=None)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran


def handler(args):
    kwa = dict(args=args)
    return [doing.doify(rollback, **kwa)]


def rollback(tymth, tock=0.0, **opts):
    """ Command line status handler

    """
    _ = (yield tock)
    args = opts["args"]
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran

    try:
        with existing.existingHby(name=name, base=base, bran=bran) as hby:
            if alias is None:
                alias = existing.aliasInput(hby)

            hab = hby.habByName(alias)

            if hab.kever.ilk not in (coring.Ilks.ixn,):
                raise kering.ValidationError(f"only interaction events can be rolled back, top event is "
                                             f"{hab.kever.ilk}")

            serder = hab.kever.serder
            dgkey = dbing.dgKey(hab.pre, serder.saidb)
            wigs = hby.db.getWigs(dgkey)

            if len(wigs) > 0:
                raise kering.ValidationError(f"top event at sequence number {hab.kever.sn} has been published to "
                                             f"{len(wigs)} witnesses, unable to rollback.")

            ked = hby.db.states.getDict(keys=serder.pre)
            pdig = hby.db.getKeLast(dbing.snKey(serder.preb, serder.sn - 1))

            pDgKey = dbing.dgKey(serder.preb, bytes(pdig))  # get message
            raw = hby.db.getEvt(key=pDgKey)
            pserder = serdering.SerderKERI(raw=bytes(raw))

            dgkey = dbing.dgKey(serder.preb, serder.saidb)
            hby.db.delEvt(dgkey)
            hby.db.wits.rem(keys=dgkey)
            hby.db.delWigs(dgkey)
            hby.db.delSigs(dgkey)  # idempotent
            hby.db.delDts(dgkey)  # idempotent do not change dts if already
            hby.db.delKes(dbing.snKey(serder.preb, serder.sn))

            seqner = coring.Number(num=serder.sn - 1)
            fner = coring.Number(numh=ked['f'])
            fner = coring.Number(num=fner.num - 1)

            # Update the only items in state that will change after rolling back an ixn
            ked['s'] = seqner.numh
            ked['et'] = pserder.ked['t']
            ked['p'] = pserder.ked['p']
            ked['d'] = pserder.said
            ked['f'] = fner.numh
            ked['dt'] = helping.nowIso8601()

            state = serdering.SerderKERI(ked=ked)  # This is wrong key state is not Serder anymore
            hby.db.states.pin(keys=hab.pre,
                              val=helping.datify(basing.KeyStateRecord,
                                                 state.ked))

            # Refresh all habs to reload this one
            hby.db.reload()
            hby.loadHabs()

            print(f"Key event at {hab.kever.sn} rolledback, current state:")
            displaying.printIdentifier(hby, hab.pre)

    except ConfigurationError as e:
        print(f"identifier prefix for {name} does not exist, rollback failed")
        return -1



# -*- encoding: utf-8 -*-
"""
keri.kli.common.config

"""
import json
from dataclasses import dataclass, asdict
from json import JSONDecodeError
from typing import Optional, List

from keri import kering


@dataclass
class Config:
    """
    Config representation of prefix configuration
    """
    name: str  # Alias for prefix
    witnessExpire: Optional[int] = 60  # timeout for witness connection
    witnesses: Optional[List[str]] = None  # list of witness nodes
    transferable: Optional[bool] = None  # Use transferable derivation code

    def __iter__(self):
        return iter(asdict(self))


def loadConfig(file: str) -> Config:
    """
        Parameters:
            file (str): file location to load from
    """
    cfg = None
    if file != '':
        with open(file, "r") as f:
            cfg = Config(**json.loads(f.read()))
    return cfg


def parseData(data_path):
    """
    Load JSON data from data path

    Parameters:
        data_path (str): path to file containing anchor/seal configuration data
    """
    try:
        if data_path.startswith("@"):
            f = open(data_path[1:], "r")
            data = json.load(f)
        else:
            data = json.loads(data_path)
    except json.JSONDecodeError:
        raise kering.ConfigurationError("data supplied to anchor in a seal must be valid JSON")

    if not isinstance(data, list):
        data = [data]

    return data


def checkRequiredArgs(args, required_args):
    """
    Ensure required arguments are present or raise error.

    Parameters:
        args (Namespace)    : the command line arguments passed in
        required_args (list): a list of strings of the names of required arguments
    """
    for required_arg in required_args:
        try:
            result = getattr(args, required_arg)
            if result is None:
                raise ValueError(f"Required arg {required_arg} not present when config file not specified")
        except AttributeError as e:
            print("all of", required_args, "must be present if not using a configuration file")
            raise ValueError(f"Required arg {required_arg} not present when config file not specified") from e


def loadFileOptions(file_path, options_class):
    """
    Load configuration JSON from the file_path and instantiate an options_class from it

    Parameters:
        file_path (str)      : the path to the configuration file to load
        options_class (class): the type of class to instantiate from the parsed JSON of the file loaded
    """
    try:
        f = open(file_path)
        config = json.load(f)

        options = options_class(**config)
    except FileNotFoundError as e:
        print("config file", file_path, "not found")
        raise e
    except JSONDecodeError as e:
        print("config file", file_path, "not valid JSON")
        raise e

    return options



import argparse
import json

from keri.app.cli.commands.multisig import incept

parser = argparse.ArgumentParser(description="Run a demo collection of witnesses")
parser.set_defaults(handler=lambda args: demo(args))
parser.add_argument('--file', '-f', help='Filename to use to create the identifier', default="", required=True)


def demo(args):
    f = open(args.file)
    opts = json.load(f)

    ms1 = incept.GroupMultisigIncept(name="multisig1", group="multigroup1", proto="http", **opts)
    ms2 = incept.GroupMultisigIncept(name="multisig2", group="multigroup2", proto="http", **opts)
    ms3 = incept.GroupMultisigIncept(name="multisig3", group="multigroup3", proto="http", **opts)

    doers = [ms1, ms2, ms3]
    return doers



# -*- encoding: utf-8 -*-
"""
keri.app.connecting module

"""
import re
import json

from ordered_set import OrderedSet as oset

from keri import kering


class Organizer:
    """ Organizes contacts relating contact information to AIDs """

    def __init__(self, hby):
        """ Create contact Organizer

        Parameters:
            hby (Habery): database environment for contact information
        """
        self.hby = hby

    def update(self, pre, data):
        """ Add or update contact information in data for the identifier prefix

        Parameters:
            pre (str): qb64 identifier prefix of contact information to update
            data (dict): data to add to or update in contact information

        """
        existing = self.get(pre)
        if existing is None:
            existing = dict()

        existing |= data

        raw = json.dumps(existing).encode("utf-8")
        cigar = self.hby.signator.sign(ser=raw)

        self.hby.db.ccigs.pin(keys=(pre,), val=cigar)
        self.hby.db.cons.pin(keys=(pre,), val=raw)

        for field, val in data.items():
            self.hby.db.cfld.pin(keys=(pre, field), val=val)

    def replace(self, pre, data):
        """ Replace all contact information for identifier prefix with data

        Parameters:
            pre (str): qb64 identifier prefix of contact information to replace
            data (dict): data to replace contact information with

        """
        self.rem(pre)
        self.update(pre, data)

    def set(self, pre, field, val):
        """ Add or replace one value in contact information for identifier prefix

        Parameters:
            pre (str): qb64 identifier prefix for contact
            field (str): field to set
            val (Union[str,bytes]): data value

        """
        data = self.get(pre) or dict()
        data[field] = val
        self.replace(pre, data)
        self.hby.db.cfld.pin(keys=(pre, field), val=val)

    def unset(self, pre, field):
        """ Remove field from contact information for identifier prefix

        Parameters:
            pre (str): qb64 identifier prefix for contact
            field (str): field to remove

        """
        data = self.get(pre)
        del data[field]
        self.replace(pre, data)
        self.hby.db.cfld.rem(keys=(pre, field))

    def rem(self, pre):
        """ Remove all contact information for identifier prefix

        Parameters:
            pre (str): qb64 identifier prefix for contact to remove

        Returns:

        """
        self.hby.db.ccigs.rem(keys=(pre,))
        self.hby.db.cons.rem(keys=(pre,))
        return self.hby.db.cfld.trim(keys=(pre,))

    def get(self, pre, field=None):
        """ Retrieve all contact information for identifier prefix

        Parameters:
            pre (str): qb64 identifier prefix for contact
            field (str): optional field name to retrieve a single field value

        Returns:
            dict: Contact data

        """
        raw = self.hby.db.cons.get(keys=(pre,))
        if raw is None:
            return None
        cigar = self.hby.db.ccigs.get(keys=(pre,))

        if not self.hby.signator.verify(ser=raw.encode("utf-8"), cigar=cigar):
            raise kering.ValidationError(f"failed signature on {pre} contact data")

        data = json.loads(raw)
        if data is None:
            return None

        if field is not None:
            return data[field] if field in data else None

        data["id"] = pre
        return data

    def list(self):
        """ Return list of all contact information for all remote identifiers

        Returns:
            list: All contact information

        """
        key = ""
        data = None
        contacts = []
        for (pre, field), val in self.hby.db.cfld.getItemIter():
            if pre != key:
                if data is not None:
                    contacts.append(data)
                data = dict(id=pre)
                key = pre

            data[field] = val

        if data is not None:
            contacts.append(data)

        return contacts

    def find(self, field, val):
        """ Find all contact information for all contacts that have the val in field

        Parameters:
            field (str): field name to search for
            val (Union[str,bytes,list]): value to search for

        Returns:
            list: All contacts that match the val in field

        """
        pres = []
        prog = re.compile(f".*{val}.*", re.I)
        for (pre, f), v in self.hby.db.cfld.getItemIter():
            if f == field and prog.match(v):
                pres.append(pre)

        return [self.get(pre) for pre in pres]

    def values(self, field, val=None):
        """ Find unique values for field in all contacts

        Args:
            field (str): field to load values for
            val (Optional(str|None): optional filter for the value of the grouped field

        Returns:
            list: Unique values from all contacts for field

        """
        prog = re.compile(f".*{val}.*", re.I) if val is not None else None

        vals = oset()
        for (pre, f), v in self.hby.db.cfld.getItemIter():
            if f == field:
                if prog is None or prog.match(v):
                    vals.add(v)

        return list(vals)

    def setImg(self, pre, typ, stream):
        """ Upload image for identifier prefix

        Streams image data in 4k chunks into database and sets content type and content length.
        Performs a full replace of all data for image of specified identifier

        Parameters:
            pre (str): qb64 identifier prefix for image
            typ (str): image content mime type
            stream (file): file-like stream of image data

        """
        self.hby.db.delTopVal(db=self.hby.db.imgs, key=pre.encode("utf-8"))

        key = f"{pre}.content-type".encode("utf-8")
        self.hby.db.setVal(db=self.hby.db.imgs, key=key, val=typ.encode("utf-8"))

        idx = 0
        size = 0
        while True:
            chunk = stream.read(4096)
            if not chunk:
                break
            key = f"{pre}.{idx}".encode("utf-8")
            self.hby.db.setVal(db=self.hby.db.imgs, key=key, val=chunk)
            idx += 1
            size += len(chunk)

        key = f"{pre}.content-length".encode("utf-8")
        self.hby.db.setVal(db=self.hby.db.imgs, key=key, val=size.to_bytes(4, "big"))

    def getImgData(self, pre):
        """ Get image metadata for identifier image if one exists

            Parameters:
                pre (str): qb64 identifier prefix for image

            Returns:
                dict: image metadata including length and type

        """
        key = f"{pre}.content-length".encode("utf-8")
        size = self.hby.db.getVal(db=self.hby.db.imgs, key=key)
        if size is None:
            return None

        key = f"{pre}.content-type".encode("utf-8")
        typ = self.hby.db.getVal(db=self.hby.db.imgs, key=key)
        if typ is None:
            return None

        return dict(
            type=bytes(typ).decode("utf-8"),
            length=int.from_bytes(size, "big")
        )

    def getImg(self, pre):
        """ Generator that yields image data in 4k chunks for identifier

        Parameters:
            pre (str): qb64 identifier prefix for image

        """
        idx = 0
        while True:
            key = f"{pre}.{idx}".encode("utf-8")
            chunk = self.hby.db.getVal(db=self.hby.db.imgs, key=key)
            if not chunk:
                break
            yield bytes(chunk)
            idx += 1



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import json

from hio import help
from hio.base import doing

from keri import kering
from keri.app import indirecting, habbing, forwarding, grouping
from keri.app.cli.common import existing
from keri.core import eventing, parsing, coring

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Add new endpoint role authorization.')
parser.set_defaults(handler=lambda args: add_end(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--aid", help="qualified base64 of AID to export rpy messages for all endpoints.",
                    required=True)


def add_end(args):
    """ Command line tool for adding endpoint role authorizations

    """
    ld = RoleDoer(name=args.name,
                  base=args.base,
                  alias=args.alias,
                  bran=args.bran,
                  aid=args.aid)
    return [ld]


class RoleDoer(doing.DoDoer):

    def __init__(self, name, base, alias, bran, aid):
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hab = self.hby.habByName(alias)
        if self.hab is None:
            raise kering.ConfigurationError(f"unknown alias={alias}")

        self.aid = aid
        doers = [doing.doify(self.roleDo)]

        super(RoleDoer, self).__init__(doers=doers)

    def roleDo(self, tymth, tock=0.0):
        """ Export any end reply messages previous saved for the provided AID

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        ends = self.hab.endsFor(self.aid)
        print(json.dumps(ends, indent=1))
        return



# -*- encoding: utf-8 -*-
"""
keri.core.routing module

"""
import datetime
import logging
import re

from hio.help import decking

from . import eventing, coring, serdering
from .. import help, kering
from ..db import dbing
from ..help import helping

logger = help.ogler.getLogger()


class Router:
    """ Reply message router

    Reply message router that accepts registration of route `r` handlers and dispatches
    reply messages to the appropriate handler.

    """

    defaultResourceFunc = "processReply"

    def __init__(self, routes=None):
        """ Initialized instance with optiona list of existing routes

        Parameters:
            routes (list): preregistered routes for this router

        """
        self.routes = routes if routes is not None else list()

    def addRoute(self, routeTemplate, resource, suffix=None):
        """ Add a route between a route template and a resource


        Parameters:
            routeTemplate (str): a route template to use for the resource
            resource (object): the resource instance to associate with the route template
            suffix(str, optional): Optional responder name suffix for this route. If a suffix is provided,
              Router will map reply routes to processReply{suffix}().  In this way, multiple closely-related routes
              can be mapped to the same resource.

        """

        fields, regex = compile_uri_template(routeTemplate)
        self.routes.append(Route(regex=regex, fields=fields, resource=resource, suffix=suffix))

    def dispatch(self, serder, saider, cigars, tsgs):
        """

        Parameters:
            serder:
            saider:
            cigars:
            tsgs:

        Returns:

        """
        ked = serder.ked
        # Dispatch based on route
        r = ked["r"]
        route, match = self._find(route=r)
        if route is None:
            raise kering.ValidationError(f"No resource is registered to handle route {r}")

        fname = self.defaultResourceFunc
        if route.suffix is not None:
            fname += route.suffix

        kwargs = match.groupdict()
        for name in route.fields:
            if name not in kwargs:
                raise kering.ValidationError(f"parameter {name} not found in route {r}")

        fn = getattr(route.resource, fname, self.processRouteNotFound)
        fn(serder=serder, saider=saider, route=r, cigars=cigars, tsgs=tsgs, **kwargs)

    def _find(self, route):
        """ Linear seach thru added routes, returning the first one that matchs

        Searches through the registered routes until a regex in one of the routes matches
        the provided route and returns the Route object along with the re.Match object.

        Parameters:
            route (str): the route from the `r` of the reply message

        Returns:
            Route: the Route object with the resource that is registered to process this rpy message
            re.Match:  the regular expression match that contains the grouping of matched parameters.

        """
        for r in self.routes:
            if res := r.regex.search(route):
                return r, res

        return None, None

    def processRouteNotFound(self, *, serder, saider, route,
                             cigars=None, tsgs=None, **kwargs):
        """ Default handler for processing reply message with an unregistered route

        Parameters:
            serder (Serder): reply event message
            saider (Saider): SAIDer of the sender
            route (str): route ('r') of the event message
            cigars (Optional(list)): list of non-transferable signature tuples
            tsgs (Optional(list)): list of transferable signature tuples
            **kwargs (dict) additional keyword args

        """

        raise kering.ConfigurationError(f"Resource registered for route {route} in {coring.Ilks.rpy}"
                                        f"does not contain the correct processReply method")


class Revery:
    """ Reply message event processor

    """

    TimeoutRPE = 3600  # seconds to timeout reply message escrows

    def __init__(self, db, rtr=None, cues=None, lax=True, local=False):
        """

        Parameters:
            db:
            cues:
            lax:
            local:
        """
        self.db = db
        self.rtr = rtr if rtr is not None else Router()
        self.cues = cues if cues is not None else decking.Deck()
        self.lax = True if lax else False  # promiscuous mode
        self.local = True if local else False  # local vs nonlocal restrictions

    @property
    def prefixes(self):
        """
        Returns .db.prefixes
        """
        return self.db.prefixes

    def processReply(self, serder, cigars=None, tsgs=None):
        """
         Process one reply message with either attached nontrans signing couples
         in cigars or attached trans indexed sig groups in tsgs. Process logic
         is route dependent and dispatched by route.

         Parameters:
             serder (Serder): instance of reply message
             cigars (list): of Cigar instances that contain nontrans signing couple
                           signature in .raw and public key in .verfer
             tsgs (list): tuples (quadruples) of form
                 (prefixer, seqner, diger, [sigers]) where:
                 prefixer is pre of trans endorser
                 seqner is sequence number of trans endorser's est evt for keys for sigs
                 diger is digest of trans endorser's est evt for keys for sigs
                 [sigers] is list of indexed sigs from trans endorser's keys from est evt

         BADA (Best Available Data Acceptance) model for each reply message.
         Latest-Seen-Signed Pairwise comparison of new update reply compared to
         old already accepted reply from same source for same route (same data).
         Accept new reply (update) if new reply is later than old reply where:
             1) Later means date-time-stamp of new is greater than old
         If non-trans signer then also (AND)
             2) Later means sn (sequence number) of last (if forked) Est evt that
                provides keys for signature(s) of new is greater than or equal to
                sn of last Est evt that provides keys for signature(s) of new.

         If nontrans and last Est Evt is not yet accepted then escrow.
         If nontrans and partially signed then escrow.

         Escrow process logic is route dependent and is dispatched by route,
         i.e. route is address of buffer with route specific handler of escrow.
        """
        for k in eventing.RPY_LABELS:
            if k not in serder.ked:
                raise kering.ValidationError(f"Missing element={k} from {coring.Ilks.rpy}"
                                             f" msg={serder.ked}.")
        # fetch from serder to process
        ked = serder.ked

        # verify said of reply
        saider = coring.Saider(qb64=ked["d"])
        if not saider.verify(sad=ked, prefixed=True):
            raise kering.ValidationError(f"Invalid said = {saider.qb64} for reply "
                                         f"msg={ked}.")

        self.rtr.dispatch(serder=serder, saider=saider, cigars=cigars, tsgs=tsgs)

    def acceptReply(self, serder, saider, route, aid, osaider=None,
                    cigars=None, tsgs=None):
        """ Applies Best Available Data Acceptance policy to reply and signatures

        Returns:
            bool: True is successfully accepted. False otherwise

        Parameters:
            serder (Serder): instance of reply msg (SAD)
            saider (Saider): instance  from said in serder (SAD)
            osaider (Saider): instance of saider for previous reply if any
            route (str): reply route
            aid (str): identifier prefix qb64 of authorizing attributable ID
            cigars (list): of Cigar instances that contain nontrans signing couple
                          signature in .raw and public key in .verfer
            tsgs (list): tuples (quadruples) of form
                (prefixer, seqner, diger, [sigers]) where:
                prefixer is pre of trans endorser
                seqner is sequence number of trans endorser's est evt for keys for sigs
                diger is digest of trans endorser's est evt for keys for sigs
                [sigers] is list of indexed sigs from trans endorser's keys from est evt

        BADA (Best Available Data Acceptance) model for each reply message.
        Latest-Seen-Signed Pairwise comparison of new update reply compared to
        old already accepted reply from same source for same route (same data).
        Accept new reply (update) if new reply is later than old reply where:
            1) If transferable: Later is True
                 A) If sn (sequence number) of last (if forked) Est evt that provides
                 keys for signature(s) of new is greater than sn of last Est evt
                 that provides keys for signature(s) of old.

                 Or

                 B) If sn of new equals sn of old And date-time-stamp of new is
                    greater than old

            2) Else If non-transferable: Later it True
                 If date-time-stamp of new is greater than old

            4) Else Later is False


        If nontrans and last Est Evt is not yet accepted then escrow.
        If nontrans and partially signed then escrow.

        Escrow process logic is route dependent and is dispatched by route,
        i.e. route is address of buffer with route specific handler of escrow.

        """
        # BADA logic.
        accepted = False  # flag to raise UnverifiedReplyError not accepted
        cigars = cigars if cigars is not None else []
        tsgs = tsgs if tsgs is not None else []

        # Is new later than old if old?
        # get date-time raises error if empty or invalid format
        dater = coring.Dater(dts=serder.ked["dt"])
        odater = None
        if osaider:
            odater = self.db.sdts.get(keys=osaider.qb64b)

        for cigar in cigars:  # process each couple to verify sig and write to db
            if cigar.verfer.transferable:  # ignore invalid transferable verfers
                continue  # skip invalid transferable

            if not self.lax and cigar.verfer.qb64 in self.prefixes:  # own cig
                if not self.local:  # own cig when not local so ignore
                    logger.info("Kevery process: skipped own attachment"
                                " on nonlocal reply msg=\n%s\n", serder.pretty())
                    continue  # skip own cig attachment on non-local reply msg

            if aid != cigar.verfer.qb64:  # cig not by aid
                logger.info("Kevery process: skipped cig not from aid="
                            "%s on reply msg=\n%s\n", aid, serder.pretty())
                continue  # skip invalid cig's verfer is not aid

            if odater:  # get old compare datetimes to see if later
                if dater.datetime <= odater.datetime:
                    logger.info("Kevery process: skipped stale update from "
                                "%s of reply msg=\n%s\n", aid, serder.pretty())
                    continue  # skip if not later
                    # raise ValidationError(f"Stale update of {route} from {aid} "
                    # f"via {Ilks.rpy}={serder.ked}.")

            if not cigar.verfer.verify(cigar.raw, serder.raw):  # cig not verify
                logger.info("Kevery process: skipped nonverifying cig from "
                            "%s on reply msg=\n%s\n", cigar.verfer.qb64, serder.pretty())
                continue  # skip if cig not verify

            # All constraints satisfied so update
            self.updateReply(serder=serder, saider=saider, dater=dater, cigar=cigar)
            self.removeReply(saider=osaider)  # remove obsoleted reply artifacts
            accepted = True
            break  # first valid cigar sufficient ignore any duplicates in cigars

        for prefixer, seqner, ssaider, sigers in tsgs:  # iterate over each tsg
            if not self.lax and prefixer.qb64 in self.prefixes:  # own sig
                if not self.local:  # own sig when not local so ignore
                    logger.info("Kevery process: skipped own attachment"
                                " on nonlocal reply msg=\n%s\n", serder.pretty())
                    continue  # skip own sig attachment on non-local reply msg

            spre = prefixer.qb64
            if aid != spre:  # sig not by aid
                logger.info("Kevery process: skipped signature not from aid="
                            "%s on reply msg=\n%s\n", aid, serder.pretty())
                continue  # skip invalid signature is not from aid

            if osaider:  # check if later logic  sn > or sn == and dt >
                if otsgs := eventing.fetchTsgs(db=self.db.ssgs, saider=osaider):
                    _, osqr, _, _ = otsgs[0]  # zeroth should be authoritative

                    if seqner.sn < osqr.sn:  # sn earlier
                        logger.info("Kevery process: skipped stale key state sig"
                                    "from %s sn=%s<%s on reply msg=\n%s\n",
                                    aid, seqner.sn, osqr.sn, serder.pretty())
                        continue  # skip if sn earlier

                    if seqner.sn == osqr.sn:  # sn same so check datetime
                        if odater:
                            if dater.datetime <= odater.datetime:
                                logger.info("Kevery process: skipped stale key"
                                            "state sig datetime from %s on reply msg=\n%s\n",
                                            aid, serder.pretty())
                                continue  # skip if not later

            # retrieve sdig of last event at sn of signer.
            sdig = self.db.getKeLast(key=dbing.snKey(pre=spre, sn=seqner.sn))
            if sdig is None:
                # create cue here to request key state for sprefixer signer
                # signer's est event not yet in signer's KEL
                self.escrowReply(serder=serder, saider=saider, dater=dater,
                                 route=route, prefixer=prefixer, seqner=seqner,
                                 ssaider=ssaider, sigers=sigers)
                self.cues.append(dict(kin="query", q=dict(pre=spre)))
                continue

            # retrieve last event itself of signer given sdig
            sraw = self.db.getEvt(key=dbing.dgKey(pre=spre, dig=bytes(sdig)))
            # assumes db ensures that sraw must not be none because sdig was in KE
            sserder = serdering.SerderKERI(raw=bytes(sraw))
            if sserder.said != ssaider.qb64:  # signer's dig not match est evt
                raise kering.ValidationError(f"Bad trans indexed sig group at sn = "
                                             f"{seqner.sn} for reply = {serder.ked}.")
            # verify sigs
            if not (sverfers := sserder.verfers):
                raise kering.ValidationError(f"Invalid reply from signer={spre}, no "
                                             f"keys at signer's est. event sn={seqner.sn}.")

            # fetch any escrowed sigs, extract just the siger from each quad
            # want sn in numerical order so use hex
            quadkeys = (saider.qb64, prefixer.qb64, f"{seqner.sn:032x}", ssaider.qb64)
            esigers = self.db.ssgs.get(keys=quadkeys)
            sigers.extend(esigers)
            sigers, valid = eventing.validateSigs(serder=serder,
                                                  sigers=sigers,
                                                  verfers=sverfers,
                                                  tholder=sserder.tholder)
            # no error so at least one verified siger

            if valid:  # meet threshold so save
                # All constraints satisfied so update
                self.updateReply(serder=serder, saider=saider, dater=dater,
                                 prefixer=prefixer, seqner=seqner, diger=ssaider,
                                 sigers=sigers)
                self.removeReply(saider=osaider)  # remove obsoleted reply artifacts
                # remove stale signatures .ssgs for this saider
                # this ensures that zeroth tsg is authoritative
                for prr, snr, dgr, _ in eventing.fetchTsgs(db=self.db.ssgs, saider=saider, snh=seqner.snh):
                    if ((snr.sn < seqner.sn) or
                            (snr.sn == seqner.sn and dgr.qb64 != ssaider.qb64)):
                        self.db.ssgs.trim(keys=(prr.qb64, f"{snr.sn:032h}", dgr.qb64, ""))

                accepted = True

            else:  # not meet threshold so escrow
                self.escrowReply(serder=serder, saider=saider, dater=dater,
                                 route=route, prefixer=prefixer, seqner=seqner,
                                 ssaider=ssaider, sigers=sigers)

        return accepted

    def updateReply(self, *, serder, saider, dater, cigar=None, prefixer=None,
                    seqner=None, diger=None, sigers=None):
        """ Update Reply SAD in database

        Update Reply SAD in database given by by serder and associated databases
        for attached cig couple or sig quadruple.
        Overwrites val at key if already exists.

        Parameters:
            serder (Serder): instance of reply msg (SAD)
            saider (Saider): instance  from said in serder (SAD)
            dater (Dater): instance from date-time in serder (SAD)
            cigar (Cigar): instance that contain nontrans signing couple
                          signature in .raw and public key in .verfer
            prefixer (Prefixer): is pre of trans endorser
            seqner (Seqner): is sequence number of trans endorser's est evt for keys for sigs
            diger (Diger): is digest of trans endorser's est evt for keys for sigs
            sigers (list): of indexed sigs from trans endorser's key from est evt

        """
        keys = (saider.qb64,)
        self.db.sdts.put(keys=keys, val=dater)  # first one idempotent
        self.db.rpys.put(keys=keys, val=serder)  # first one idempotent
        if cigar:
            self.db.scgs.put(keys=keys, vals=[(cigar.verfer, cigar)])
        if sigers:  # want sn in numerical order so use hex
            quadkeys = (saider.qb64, prefixer.qb64, f"{seqner.sn:032x}", diger.qb64)
            self.db.ssgs.put(keys=quadkeys, vals=sigers)

    def removeReply(self, saider):
        """ Remove Reply SAD artifacts given by saider.

        Parameters:
            saider (Saider): instance from said in serder (SAD)

        """
        if saider:
            keys = (saider.qb64,)

            self.db.ssgs.trim(keys=(saider.qb64, ""))  # remove whole branch
            self.db.scgs.rem(keys=keys)
            self.db.rpys.rem(keys=keys)
            self.db.sdts.rem(keys=keys)

    def escrowReply(self, *, serder, saider, dater, route, prefixer, seqner,
                    ssaider, sigers):
        """ Escrow reply by route

        Parameters:
            serder (Serder): instance of reply msg (SAD)
            saider (Saider): instance  from said in serder (SAD)
            dater (Dater): instance from date-time in serder (SAD)
            route (str): reply route
            prefixer (Prefixer): is pre of trans endorser
            seqner (Seqner): is sequence number of trans endorser's est evt for keys for sigs
            ssaider (Saider) is said of trans endorser's est evt for keys for sigs
            sigers (list): is indexed sigs from trans endorser's key from est evt

        """
        if not sigers:
            return  # nothing to escrow
        keys = (saider.qb64,)
        self.db.sdts.put(keys=keys, val=dater)  # first one idempotent
        self.db.rpys.put(keys=keys, val=serder)  # first one idempotent
        quadkeys = (saider.qb64, prefixer.qb64, f"{seqner.sn:032x}", ssaider.qb64)
        self.db.ssgs.put(keys=quadkeys, vals=sigers)
        self.db.rpes.put(keys=(route,), vals=[saider])

    def processEscrowReply(self):
        """ Process escrows for reply messages.

        Escrows are keyed by reply route and val is reply said

        triple (prefixer, seqner, diger)
        quadruple (prefixer, seqner, diger, siger)

        """
        for (route, ion), saider in self.db.rpes.getIoItemIter():
            try:
                tsgs = eventing.fetchTsgs(db=self.db.ssgs, saider=saider)

                keys = (saider.qb64,)
                dater = self.db.sdts.get(keys=keys)
                serder = self.db.rpys.get(keys=keys)
                try:
                    if not (dater and serder and tsgs):
                        raise ValueError(f"Missing escrow artifacts at said={saider.qb64}"
                                         f"for route={route}.")

                    # do date math for stale escrow
                    if ((helping.nowUTC() - dater.datetime) >
                            datetime.timedelta(seconds=self.TimeoutRPE)):
                        # escrow stale so raise ValidationError which unescrows below
                        logger.info("Kevery unescrow error: Stale reply escrow "
                                    " at route = %s\n", route)

                        raise kering.ValidationError(f"Stale reply escrow at route = {route}.")

                    self.processReply(serder=serder, tsgs=tsgs)

                except kering.UnverifiedReplyError as ex:
                    # still waiting on missing prior event to validate
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception("Kevery unescrow attempt failed: %s\n", ex.args[0])
                    else:
                        logger.error("Kevery unescrow attempt failed: %s\n", ex.args[0])

                except Exception as ex:  # other error so remove from reply escrow
                    self.db.rpes.rem(keys=(route, ), val=saider)  # remove escrow only
                    self.removeReply(saider)  # remove escrow reply artifacts
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception("Kevery unescrowed due to error: %s\n", ex.args[0])
                    else:
                        logger.error("Kevery unescrowed due to error: %s\n", ex.args[0])

                else:  # unescrow succeded
                    self.db.rpes.rem(keys=(route, ), val=saider)  # remove escrow only
                    logger.info("Kevery unescrow succeeded for reply=\n%s\n",
                                serder.pretty())

            except Exception as ex:  # log diagnostics errors etc
                self.db.rpes.rem(keys=(route,), val=saider)  # remove escrow only
                self.removeReply(saider)  # remove escrow reply artifacts
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Kevery unescrowed due to error: %s\n", ex.args[0])
                else:
                    logger.error("Kevery unescrowed due to error: %s\n", ex.args[0])


class Route:
    """ Route class for registration of reply message handlers

    This class represents a registered route internally to the Revery.
    the properties are created by using the Falcon compile route utility method

    Properties:
        .regex(re): compiled url template regex
        .fields(set): field names for matches in regex
        .resource(object): the handler for this route
        .suffix(Optional(str)): a suffix to be applied to the handler method

    """

    def __init__(self, regex, fields, resource, suffix=None):
        """ Initialize instance of route

        Parameters:
            regex(re): compiled url template regex
            fields(set): field names for matches in regex
            resource(object): the handler for this route
            suffix(Optional(str)): a suffix to be applied to the handler method

        """
        self.regex = regex
        self.fields = fields
        self.resource = resource
        self.suffix = suffix


def compile_uri_template(template):
    """Compile the given URI template string into a pattern matcher.

    This function can be used to construct custom routing engines that
    iterate through a list of possible routes, attempting to match
    an incoming request against each route's compiled regular expression.

    Each field is converted to a named group, so that when a match
    is found, the fields can be easily extracted using
    :py:meth:`re.MatchObject.groupdict`.

    This function does not support the more flexible templating
    syntax used in the default router. Only simple paths with bracketed
    field expressions are recognized. For example::

        /
        /books
        /books/{isbn}
        /books/{isbn}/characters
        /books/{isbn}/characters/{name}

    Also, note that if the template contains a trailing slash character,
    it will be stripped in order to normalize the routing logic.

    Args:
        template(str): The template to compile. Note that field names are
            restricted to ASCII a-z, A-Z, and the underscore character.

    Returns:
        tuple: (template_field_names, template_regex)
    """

    if not isinstance(template, str):
        raise TypeError('uri_template is not a string')

    if not template.startswith('/'):
        raise ValueError("uri_template must start with '/'")

    if '//' in template:
        raise ValueError("uri_template may not contain '//'")

    if template != '/' and template.endswith('/'):
        template = template[:-1]

    # template names should be able to start with A-Za-z
    # but also contain 0-9_ in the remaining portion
    expression_pattern = r'{([a-zA-Z]\w*)}'

    # Get a list of field names
    fields = set(re.findall(expression_pattern, template))

    # Convert Level 1 var patterns to equivalent named regex groups
    escaped = re.sub(r'[\.\(\)\[\]\?\*\+\^\|]', r'\\\g<0>', template)
    pattern = re.sub(expression_pattern, r'(?P<\1>[^/]+)', escaped)
    pattern = r'\A' + pattern + r'\Z'

    return fields, re.compile(pattern, re.IGNORECASE)



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands.ipex module

Join multisig ipex messages
"""
import argparse

from hio.base import doing

from keri import help
from keri.app import habbing, indirecting, agenting, notifying, grouping, connecting, forwarding
from keri.app.cli.common import existing
from keri.core import parsing, routing, serdering, coring
from keri.peer import exchanging
from keri.vdr import verifying, credentialing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Join group multisig ipex events')
parser.set_defaults(handler=lambda args: join(args))
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--auto", "-Y", help="auto approve any delegation request non-interactively", action="store_true")


def join(args):
    """  Wait for and provide interactive confirmation of group multisig inception, rotation or interaction events

    Parameters:
        args(Namespace): parsed arguements namespace object

    """
    name = args.name
    base = args.base
    bran = args.bran
    auto = args.auto

    joinDoer = JoinDoer(name=name, base=base, bran=bran, auto=auto)

    doers = [joinDoer]
    return doers


class JoinDoer(doing.DoDoer):
    """  Doist doer capable of polling for group ipex events and prompting user for action

    """

    def __init__(self, name, base, bran, auto=False):
        """ Create doer for polling for group ipex events and either approve automatically or prompt user

        Parameters:
            name (str): database environment name
            base (str): database directory prefix
            bran (str): passcode to unlock keystore

        """
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.rgy = credentialing.Regery(hby=self.hby, name=name, base=base)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        self.witq = agenting.WitnessInquisitor(hby=self.hby)
        self.org = connecting.Organizer(hby=self.hby)
        self.notifier = notifying.Notifier(hby=self.hby)
        self.exc = exchanging.Exchanger(hby=self.hby, handlers=[])
        self.verifier = verifying.Verifier(hby=self.hby, reger=self.rgy.reger)
        self.rvy = routing.Revery(db=self.hby.db,  lax=True)
        self.hby.kvy.registerReplyRoutes(self.rvy.rtr)
        self.psr = parsing.Parser(kvy=self.hby.kvy, tvy=self.rgy.tvy, rvy=self.rvy, vry=self.verifier, exc=self.exc)

        mux = grouping.Multiplexor(hby=self.hby, notifier=self.notifier)
        grouping.loadHandlers(exc=self.exc, mux=mux)
        self.counselor = grouping.Counselor(hby=self.hby)

        self.registrar = credentialing.Registrar(hby=self.hby, rgy=self.rgy, counselor=self.counselor)
        self.credentialer = credentialing.Credentialer(hby=self.hby, rgy=self.rgy, registrar=self.registrar,
                                                       verifier=self.verifier)

        self.mbx = indirecting.MailboxDirector(hby=self.hby, exc=self.exc, topics=['/receipt', '/multisig', '/replay',
                                                                                   '/delegate'])
        self.postman = forwarding.Poster(hby=self.hby)

        doers = [self.hbyDoer, self.witq,  self.mbx, self.counselor, self.registrar, self.credentialer, self.postman]
        self.toRemove = list(doers)
        doers.extend([doing.doify(self.joinDo)])
        self.auto = auto
        super(JoinDoer, self).__init__(doers=doers)

    def joinDo(self, tymth, tock=0.0):
        """
        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        print("Waiting for group ipex events...")

        while True:

            found = False
            for keys, notice in self.notifier.noter.notes.getItemIter():
                attrs = notice.attrs
                route = attrs['r']

                if route == '/multisig/exn':
                    said = attrs["d"]
                    exn, pathed = exchanging.cloneMessage(self.hby, said=said)
                    embeds = exn.ked['e']

                    if embeds['exn']['r'].startswith("/ipex"):
                        done = yield from self.ipex(exn, pathed)

                        if done:
                            self.notifier.noter.notes.rem(keys=keys)

                        else:
                            delete = input(f"\nDelete event [Y|n]? ")
                            if delete in ("Y", "y"):
                                self.notifier.noter.notes.rem(keys=keys)
                        found = True
            if found:
                break

            yield self.tock

        self.remove(self.toRemove)

    def ipex(self, exn, pathed):
        """  Handle exn messages for ipex

        Parameters:
            exn (SerderKERI): exn message
            pathed (dict): pathed attachments dict

        Returns:

        """
        embeds = exn.ked['e']
        sender = exn.ked['i']

        contact = self.org.get(sender)
        senderAlias = contact['alias']

        eexn = embeds['exn']

        route = eexn['r']

        group = eexn["i"]
        hab = self.hby.habs[group] if group in self.hby.habs else None
        if hab is None:
            raise ValueError(f"message sender not a valid AID={group}")

        print()
        print(f"Group IPEX Message proposal (from {senderAlias}):")
        print(f"    Message Type: {eexn['r']}")
        print(f"    Message SAID: {eexn['d']}")
        print(f"    Sending From: {hab.name} ({hab.pre})")

        match route:
            case "/ipex/admit":
                recp = yield from self.getAdmitRecp(eexn)
            case "/ipex/agree":
                recp = self.getAgreeRecp(eexn)
            case "/ipex/apply":
                recp = self.getApplyRecp(eexn)
            case "/ipex/grant":
                recp = self.getGrantRecp(eexn)
            case "/ipex/offer":
                recp = self.getOfferRecp(eexn)
            case "/ipex/spurn":
                recp = self.getSpurnRecp(eexn)
            case _:
                return False

        contact = self.org.get(recp)
        if contact is not None and "alias" in contact:
            print(f"    Sending To: {contact['alias']} ({recp})")
        else:
            print(f"    Sending To: Unknown AID ({recp})")

        if self.auto:
            approve = True
        else:
            yn = input(f"\nApprove [Y|n]? ")
            approve = yn in ('', 'y', 'Y')

        if approve:
            eserder = serdering.SerderKERI(sad=eexn)
            anc = bytearray(eserder.raw) + pathed["exn"]
            self.psr.parseOne(ims=bytes(anc))

            msg = hab.endorse(serder=eserder, last=False, pipelined=False)
            msg = msg + pathed["exn"]
            self.psr.parseOne(ims=bytes(msg))

            smids = hab.db.signingMembers(pre=hab.pre)
            smids.remove(hab.mhab.pre)

            for smid in smids:  # this goes to other participants only as a signaling mechanism
                rexn, atc = grouping.multisigExn(ghab=hab, exn=msg)
                self.postman.send(src=hab.mhab.pre,
                                  dest=smid,
                                  topic="multisig",
                                  serder=rexn,
                                  attachment=atc)
                while not self.postman.sent(said=rexn.said):
                    yield self.tock

            while not self.exc.complete(said=eserder.said):
                self.exc.processEscrow()
                yield self.tock

            if self.exc.lead(hab, said=exn.said):
                print(f"Sending message {eserder.said} to {recp}")
                atc = exchanging.serializeMessage(self.hby, eserder.said)
                del atc[:eserder.size]
                self.postman.send(src=hab.mhab.pre,
                                  dest=recp,
                                  topic="credential",
                                  serder=eserder,
                                  attachment=atc)

                while not self.postman.sent(said=eserder.said):
                    yield self.tock

                print("... ipex message sent")
                return True

            return True
        return False

    def getAdmitRecp(self, exn):
        grant, pathed = exchanging.cloneMessage(self.hby, exn['p'])
        if grant is None:
            raise ValueError(f"exn message said={exn['p']} not found")

        embeds = grant.ked['e']
        acdc = embeds["acdc"]
        for label in ("anc", "iss", "acdc"):
            ked = embeds[label]
            sadder = coring.Sadder(ked=ked)
            ims = bytearray(sadder.raw) + pathed[label]
            self.psr.parseOne(ims=ims)

        said = acdc["d"]
        while not self.rgy.reger.saved.get(keys=said):
            yield self.tock

        return acdc['i']

    def getAgreeRecp(self, exn):
        pass

    def getApplyRecp(self, exn):
        return exn['a']['i']

    def getGrantRecp(self, exn):
        return exn['a']['i']

    def getOfferRecp(self, exn):
        pass

    def getSpurnRecp(self, exn):
        pass



# -*- encoding: utf-8 -*-
"""
keri.vc.challenging module

"""

from keri.core import coring


def loadHandlers(db, signaler, exc):
    """ Load handlers for the peer-to-peer challenge response protocol

    Parameters:
        db (Baser): database environment
        signaler (Signaler): Signaler for transient messages for the controller of the agent
        exc (Exchanger): Peer-to-peer message router

    """
    chacha = ChallengeHandler(db=db, signaler=signaler)
    exc.addHandler(chacha)


class ChallengeHandler:
    """  Handle challenge response peer to peer `exn` message """

    resource = "/challenge/response"

    def __init__(self, db, signaler):
        """ Initialize peer to peer challenge response messsage """

        self.db = db
        self.signaler = signaler
        super(ChallengeHandler, self).__init__()

    def handle(self, serder, attachments=None):
        """  Do route specific processsing of Challenge response messages

        Parameters:
            serder (Serder): Serder of the exn challenge response message
            attachments (list): list of tuples of pather, CESR SAD path attachments to the exn event

        """
        payload = serder.ked['a']
        signer = serder.pre
        words = payload["words"]

        msg = dict(
            signer=signer,
            said=serder.said,
            words=words
        )

        # Notify controller of sucessful challenge
        self.signaler.push(msg, topic="/challenge")

        # Log signer against event to track successful challenges with signed response
        self.db.reps.add(keys=(signer,), val=coring.Saider(qb64=serder.said))



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import os
import stat
from pathlib import Path

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from hio import help
from hio.base import doing

from keri.app.cli.common import existing
from keri.kering import ConfigurationError

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Export keys of specified identifier for use with SSH')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', default=None)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran

parser.add_argument("--private", help="export private key instead of public key", action="store_true")
parser.add_argument("--username", help="override file name for the key to export", default=None)


def handler(args):
    kwa = dict(args=args)
    return [doing.doify(export, **kwa)]


def export(tymth, tock=0.0, **opts):
    """ Command line status handler

    """
    _ = (yield tock)
    args = opts["args"]
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran
    private = args.private
    filename = args.username if args.username else alias
    home = str(Path.home())

    try:
        with existing.existingHby(name=name, base=base, bran=bran) as hby:
            if alias is None:
                alias = existing.aliasInput(hby)

            hab = hby.habByName(alias)
            if private:
                signer = hab.ks.pris.get(hab.kever.verfers[0].qb64,
                                         decrypter=hab.mgr.decrypter)
                sigkey = ed25519.Ed25519PrivateKey.from_private_bytes(signer.raw)
                pem = sigkey.private_bytes(encoding=serialization.Encoding.PEM,
                                           format=serialization.PrivateFormat.OpenSSH,
                                           encryption_algorithm=serialization.NoEncryption())

                f = open(os.path.join(home, ".ssh", filename), "w")
                for line in pem.splitlines(keepends=True):
                    f.write(line.decode("utf-8"))
                f.close()
                os.chmod(os.path.join(home, ".ssh", filename), stat.S_IRUSR | stat.S_IWUSR)

            else:
                verkey = ed25519.Ed25519PublicKey.from_public_bytes(hab.kever.verfers[0].raw)
                pem = verkey.public_bytes(encoding=serialization.Encoding.OpenSSH,
                                          format=serialization.PublicFormat.OpenSSH)

                f = open(os.path.join(home, ".ssh", f"{filename}.pub"), "w")
                for line in pem.splitlines(keepends=True):
                    f.write(line.decode("utf-8"))


    except ConfigurationError as e:
        print(f"identifier prefix for {name} does not exist, incept must be run first", )
        return -1



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import json

from hio import help
from hio.base import doing

from keri.core import coring

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Saidify a JSON file.')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--file', '-f', help='Filename to use to create the identifier', default="", required=True)
parser.add_argument('--label', '-l', help='Field label to SAID-ify', default="d", required=False)


def handler(args):
    """
    Saidify the provided SAD

    Args:
        args(Namespace): arguments object from command line
    """
    kwa = dict(args=args)
    return [doing.doify(saidify, **kwa)]


def saidify(tock=0.0, **opts):
    _ = (yield tock)

    args = opts["args"]

    with open(args.file, 'r') as f:
        sad = json.load(f)
        _, out = coring.Saider.saidify(sad=sad, label=args.label)

    with open(args.file, 'w') as f:
        json.dump(out, f)



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import json
from ordered_set import OrderedSet as oset

from hio.base import doing

from keri import kering
from ..common import existing
from ... import habbing, agenting, indirecting

parser = argparse.ArgumentParser(description='Create and publish an interaction event')
parser.set_defaults(handler=lambda args: interact(args))
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--data', '-d', help='Anchor data, \'@\' allowed', default=None, action="store", required=False)


def interact(args):
    """
    Performs a rotation of the identifier of the environment represented by the provided name parameter

        args (parseargs):  Command line argument

    """
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran

    if args.data is not None:
        try:
            if args.data.startswith("@"):
                f = open(args.data[1:], "r")
                data = json.load(f)
            else:
                data = json.loads(args.data)
        except json.JSONDecodeError:
            raise kering.ConfigurationError("data supplied must be value JSON to anchor in a seal")

        if not isinstance(data, list):
            data = [data]

    else:
        data = None

    ixnDoer = InteractDoer(name=name, base=base, alias=alias, bran=bran, data=data)

    return [ixnDoer]


class InteractDoer(doing.DoDoer):
    """
    DoDoer that launches Doers needed to create an interaction event and publication of the event
    to all appropriate witnesses
    """

    def __init__(self, name, base, bran, alias, data: list = None):
        """
        Returns DoDoer with all registered Doers needed to perform interaction event.

        Parameters:
            name is human readable str of identifier
            proto is tcp or http method for communicating with Witness
            data is list of dicts of committed data such as seals
       """

        self.alias = alias
        self.data = data

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        self.mbx = indirecting.MailboxDirector(hby=self.hby, topics=['/receipt', "/replay", "/reply"])
        doers = [self.hbyDoer, self.mbx, doing.doify(self.interactDo)]

        super(InteractDoer, self).__init__(doers=doers)

    def interactDo(self, tymth, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        hab = self.hby.habByName(name=self.alias)
        hab.interact(data=self.data)

        witDoer = agenting.WitnessReceiptor(hby=self.hby)
        self.extend(doers=[witDoer])

        if hab.kever.wits:
            witDoer.msgs.append(dict(pre=hab.pre))
            while not witDoer.cues:
                _ = yield self.tock

        print(f'Prefix  {hab.pre}')
        print(f'New Sequence No.  {hab.kever.sn}')
        for idx, verfer in enumerate(hab.kever.verfers):
            print(f'\tPublic key {idx+1}:  {verfer.qb64}')

        toRemove = [self.hbyDoer, witDoer, self.mbx]
        self.remove(toRemove)

        return



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import json

from hio import help
from hio.base import doing

from keri.app import connecting
from keri.app.cli.common import existing
from keri.kering import ConfigurationError

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='List existing contacts')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran


def handler(args):
    kwa = dict(args=args)
    return [doing.doify(list, **kwa)]


def list(tymth, tock=0.0, **opts):
    """ Command line status handler

    """
    _ = (yield tock)
    args = opts["args"]
    name = args.name
    base = args.base
    bran = args.bran

    try:
        with existing.existingHby(name=name, base=base, bran=bran) as hby:
            org = connecting.Organizer(hby=hby)
            for c in org.list():

                aid = c['id']
                accepted = [saider.qb64 for saider in hby.db.chas.get(keys=(aid,))]
                received = [saider.qb64 for saider in hby.db.reps.get(keys=(aid,))]
                valid = set(accepted) & set(received)

                challenges = []
                for said in valid:
                    exn = hby.db.exns.get(keys=(said,))
                    challenges.append(dict(dt=exn.ked['dt'], words=exn.ked['a']['words']))

                c["challenges"] = challenges

                wellKnowns = []
                wkans = hby.db.wkas.get(keys=(aid,))
                for wkan in wkans:
                    wellKnowns.append(dict(url=wkan.url, dt=wkan.dt))

                c["wellKnowns"] = wellKnowns

                print(json.dumps(c, indent=2))

    except ConfigurationError as e:
        print(f"identifier prefix for {name} does not exist, incept must be run first", )
        return -1
